// Vector.h
//
// ICS 65 Fall 2012
// Code Example
//
// This header file declares a template class called Vector, which stores a
// dynamically-sized collection of elements of some type.  The template class
// takes one template parameter, T, which specifies the type of elements
// stored in a particular vector.

#ifndef VECTOR_H
#define VECTOR_H

#include <algorithm>



// When we want to write a template, we begin with the word "template" and
// follow it with template parameters surrounded by angle brackets.  In
// this case, we want a template class, so we put the "template <...>" in
// front of the word "class" (and, traditionally, we write it on a separate
// line, though this is not a requirement in C++).
//
// It should be noted that it is also legal to say this:
//
//     template <class T>
//
// In that context, "class" and "typename" have the same meaning; they are
// a stand-in for any type you'd like, whether it's defined by a class or
// not.  "class" is somewhat old-school; the "typename" keyword was added
// to C++ in the first standardization process in the mid-to-late 1990s,
// and is now the typical way to specify a template.

template <typename T>
class Vector
{
public:
	static const unsigned int DEFAULT_VECTOR_SIZE = 10;


	// The member functions of Vector are mostly the same as they were in
	// our non-template version; the difference is that we've replaced all
	// of the specific uses of "std::string" with "T" instead, to make clear
	// that, for a given kind of Vector, if you call at() on that Vector,
	// you get whatever kind of element is supposed to be stored in that
	// Vector.

	Vector();
	Vector(const Vector& v);
	~Vector();
	Vector& operator=(const Vector& v);

	T& at(unsigned int index);
	const T& at(unsigned int index) const;

	// One interesting tidbit is going on here.  We're passing the T parameter
	// to this member function by const reference.  The reason is that we don't
	// know the type of T -- it might be something very small and cheap to
	// copy, but it might be something much bigger and more expensive to copy.
	// To err on the side of caution, we'll declare our function to take a
	// "const T&", so if T is an expensive object to copy, it won't be copied
	// on the way into our function.
	void add(const T& t);

	unsigned int size() const;
	unsigned int capacity() const;

private:
	T* items;
	unsigned int sz;
	unsigned int cap;
};



// As in the previous version of this code example, it would be handy to
// have a utility function that copies values from one array to another.
// This is the same function we wrote previously, except that it's now
// a template function, because we need a different version of it for
// every different kind of Vector.  I've left it in the unnamed namespace,
// though, so it won't be available between source files (as it's only
// necessary in the context of using a Vector, in which case Vector.h
// will always have been included).
namespace
{
	template <typename T>
	void arrayCopy(T* target, T* source, unsigned int size)
	{
		for (unsigned int i = 0; i < size; i++)
		{
			target[i] = source[i];
		}
	}
}



// A class declaration contains, at minimum, a list of all of the members
// of the class, so that the compiler will know what variables are contained
// in objects of that class, and will also know what functions can be called
// on objects of that class.
//
// The template class declaration above plays much the same role, but remember
// that it declares a *set* of possible classes.  Still, the idea is the same:
// a template class declaration, at minimum, lists all of the members that
// should be in all of the classes in the set, in terms of the type parameters
// (T, in this example) whenever necessary.
//
// Now we need an analogue for the definitions of the class' member functions.
// The problem is this: C++ doesn't actually compile a template function -- or
// a member function of a template class -- until it is instantiated.  So,
// until you declare a Vector<int> variable somewhere, there's no such class as
// Vector<int>.  But, at the point where you declare a Vector<int>, a C++
// compiler will now need to stop and compile the Vector<int> class declaration,
// as well as any of the member functions that are subsequently called on that
// Vector<int> variable.  In order to compile them wherever used, the full
// source code needs to be made available in the header file.  So, we'll need
// to define all of the member functions right here in the header file.



// One more thing to think about: how do we implement the member functions?
// A template class is a blueprint for a set of classes.  When you declare
// a member function in a template class, you're actually declaring a set of
// functions, one for each instantiation of the template class.  So, when you
// declare a constructor in the Vector template above, you've actually declared
// a set of Vector constructors.  What we need now is a definition for the
// whole set of Vector constructors.  The good news is that we already know
// how to do that: make a template function that specifies all of the possible
// Vector constructors at once.
//
// Notice the syntax below; it's a bit tricky.  First of all, since we're
// declaring a template function, we start with the "template" keyword and a
// list of the type parameters.  Then we specify the return type (but a
// constructor has no return type, so I've left it out), then the name of
// the function.  The name of the function is listed as "Vector<T>::Vector".
// The reason for this bizarre piece of syntax is this: you need to name the
// class "Vector<T>", because that's the difference between the functions
// in the set.  But, once you've said "Vector<T>::", that puts you in the
// scope of Vector<T>, so any subsequent use of the word "Vector" is assumed
// to be "Vector<T>".


template <typename T>
Vector<T>::Vector()
	: items(new T[Vector::DEFAULT_VECTOR_SIZE]), sz(0), cap(Vector::DEFAULT_VECTOR_SIZE)
{
}


template <typename T>
Vector<T>::Vector(const Vector& v)
	: items(new T[v.cap]), sz(v.sz), cap(v.cap)
{
	arrayCopy(items, v.items, sz);
}


template <typename T>
Vector<T>::~Vector()
{
	delete[] items;
}


// There's a funny bit of syntax here that I should point out:
//
// * The return type is listed as Vector<T>
// * The function's name is listed as Vector<T>::operator=
// * The parameter's type is listed as Vector, without the <T>
//
// The reason for this is a matter of scoping.  In the scope of a class
// Vector<T>, Vector (without a type parameter) is taken to mean
// Vector<T>.  But we're only inside the scope of Vector<T> beyond where
// the "::" operator appears in the function's signature.  Before that,
// if we want to talk about a Vector of T's, we have to say so.  This
// is why we have to say "Vector<T>" for the return type (because the
// return type appears before the "::") and why we don't have to say it
// for the parameter's type (because the parameter appears after the "::").

template <typename T>
Vector<T>& Vector<T>::operator=(const Vector& v)
{
	if (this != &v)
	{
		T* newItems = new T[v.cap];
		arrayCopy(newItems, v.items, v.sz);

		sz = v.sz;
		cap = v.cap;

		delete[] items;
		items = newItems;
	}

	return *this;
}


template <typename T>
T& Vector<T>::at(unsigned int index)
{
	return items[index];
}


template <typename T>
const T& Vector<T>::at(unsigned int index) const
{
	return items[index];
}


template <typename T>
void Vector<T>::add(const T& t)
{
	if (sz == cap)
	{
		int newCap = cap * 2 + 1;

		T* newItems = new T[newCap];
		arrayCopy(newItems, items, sz);

		cap = newCap;

		delete[] items;
		items = newItems;
	}

	items[sz] = t;
	sz++;
}


template <typename T>
unsigned int Vector<T>::size() const
{
	return sz;
}


template <typename T>
unsigned int Vector<T>::capacity() const
{
	return cap;
}



// Now that we've seen all the code for all of the member functions, we can
// think a little bit about types T is allowed to be.  In general, C++ allows
// a template's type parameters to be any types you'd like; however, in
// practice, the things you to do with those types limits what the types are
// allowed to be.  So if we want to know what T is allowed to be in this
// template class, we need to consider what we're doing to T's within that
// template class.
//
// Here are the things we do to T's (and where we do them):
//
// * A few of the member functions either pass T's by reference or return T's
//   by reference.  This doesn't introduce any constraint on T, though, because
//   any type can do this.
//
// * A few of the member functions either pass T's by reference or return T's
//   by reference.  This doesn't introduce any constraint on T, though, because
//   any type can do this.
//
// * The member variable "items" is a pointer to a T.  This also introduces no
//   meaningful constraint on T, because you can have a pointer to any type of
//   value.
//
// * The constructor, copy constructor, assignment operator, and the add()
//   function all create a dynamically-allocated array of T's (i.e., they
//   all say "new T[...]").  This imposes a limitation on T: creating an
//   array of T's requires default-constructing T objects, which means that
//   T's must have a constructor that takes no parameters.
//
// * The copy constructor, assignment operator, and the add() function all
//   copy a sequence of T's from one array to another using the std::copy_n
//   function from the standard library.  This function requires the ability
//   to copy T's from one place to another, meaning that T must have an
//   assignment operator.  (Note that all classes have assignment operators
//   even if you don't write one, unless you explicitly declare it to be
//   private.)
//
// Another thing to be aware of is that the constraints are imposed on a
// member-function-by-member-function basis.  If there was a class X and you
// said this in a source file:
//
//      Vector<X> myVector;
//
// this would compile only the class declaration for Vector<X>, its
// constructor (because declaring a Vector<X> variable constructs one)
// and its destructor (because that variable would die automatically when
// it falls out of scope).  No other member functions would be compiled
// unless you called them, and the constraints imposed by those member
// functions would only be imposed in that case.  So, for example, the
// restriction that X must have an assignment operator will only apply
// if you call a copy constructor, an assignment operator, or the add()
// function on a Vector<X>; if not, you'd have no problems.
//
// This is, in general, good news from a flexibility standpoint; you only
// face the constraints that will explicitly cause a problem in your code.
// On the other hand, this can also be a ticking time bomb; you can be
// blissfully unaware of a constraint problem early on in the development
// of a program, only to discover much later that you now need to call a
// function that imposes a constraint you weren't aware of; suddenly,
// you may have to change your design to work around the problem.



#endif // VECTOR_H
